/*
 * Copyright (c) 2020 RTSoft. All rights reserved.
 *
 * SPDX-License-Identifier: Apache-2.0
 *
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 *     http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */

#include <stdio.h>
#include <unistd.h>
#include <stdint.h>
#include <stdlib.h>
#include <string.h>
#include <errno.h>
#include <fcntl.h>
#include <sys/types.h>
#include <sys/mman.h>

static char device_name[18];

/* From txt2-rt-core.git/txt2-M4-rt-core/cubemx/txt2/Drivers/CMSIS/Device/ST/STM32MP1xx/Include/stm32mp157cxx_cm4.h */

#define UID_BASE                  (0x5C005234L)            /*!< Unique Device ID register base address */

/* From txt2-rt-core.git/txt2-M4-rt-core/ftsrc/ftserver.c */

static const char alnum[] = {
    "ABCDEFGHIJKLMNOPQRSTUVWXYZ"
    "abcdefghijklmnopqrstuvwxyz"
    "0123456789"
};

/**
 * @brief Procedure to generate a name for discovery
 */
static void GenerateShortName(uint32_t uid)
{
    int sz = sizeof(alnum) - 1; /* Must be 62 */
    /* uint32_t uid = *(uint32_t *)UID_BASE; */

    memset(device_name, 0, sizeof(device_name));
#if 1 /* 4 characters name, can produce same name for different controllers */
    uid %= sz * sz * sz * sz - 1;
    device_name[0] = alnum[uid % sz];
    uid /= sz;
    device_name[1] = alnum[uid % sz];
    uid /= sz;
    device_name[2] = alnum[uid % sz];
    uid /= sz;
    device_name[3] = alnum[uid % sz];
#else /* Up to 6 characters name, unique for controller */
    int i;
    if ((uid & 0xff000000) == 0) { /* 3-5 */
        if ((uid & 0xff00) == 0) { /* 3 */
            uid = ((uid >> 16) & 0xff00) | (uid & 0xff);
            device_name[2] = alnum[uid % sz];
            uid /= sz;
            device_name[1] = alnum[uid % sz];
            uid /= sz;
            device_name[0] = 'f' + uid;
        } else { /* 5 */
            for (i = 4;i > 0; i--) {
                device_name[i] = alnum[uid % sz];
                uid /= sz;
            }
            device_name[0] = 'x' + uid;
        }

    } else { /* Full case 6 */
        for (i = 5; i > 0; i--) {
            device_name[i] = alnum[uid % sz];
            uid /= sz;
        }
        device_name[0] = 'a' + uid;
    }
#endif
}

uint32_t read_board_uid()
{
    uint32_t buid;

    int rfid = open("/sys/bus/nvmem/devices/stm32-romem0/nvmem", O_RDONLY);

    if (rfid < 0) {
        fprintf(stderr, "Failed to open file  /sys/bus/nvmem/devices/stm32-romem0/nvmem for reading: %d %s\n",
                errno, strerror(errno));
        exit(errno);
    }

    if (lseek(rfid, 60 * 4, SEEK_SET) == -1) {
        fprintf(stderr, "Failed to seek in file /sys/bus/nvmem/devices/stm32-romem0/nvmem: %d %s\n",
            errno, strerror(errno));
        exit(errno);
    }

    if (read(rfid, &buid, sizeof(buid)) != sizeof(buid)) {
        fprintf(stderr, "Failed to read file  /sys/bus/nvmem/devices/stm32-romem0/nvmem: %d %s\n",
            errno, strerror(errno));
        exit(errno);
    }

    close(rfid);

    return buid;
}

int main(int argc, char **argv)
{
    int fd;

    if ((argc == 3 && strcmp(argv[1], "-o")) || (argc != 3 && argc != 1)) {
        fprintf(stderr, "Usage: %s [-o <file>]\n", argv[0]);
        exit(EINVAL);
    }

    GenerateShortName(read_board_uid());

    if (argc == 3) {
        fd = creat(argv[2], S_IRUSR | S_IWUSR | S_IRGRP | S_IROTH);
        if (fd == -1) {
            fprintf(stderr, "Failed to open file %s for writing: %d %s\n",
                    argv[2], errno, strerror(errno));
            exit(errno);
        }
        strcat(device_name, "\n");
        write(fd, device_name, strlen(device_name));
        close(fd);
    } else {
        printf("%s\n", device_name);
        fflush(stdout);
    }

    return 0;
}
